---------------------------------------------
-- SimpleMeter - Easy-to-use DPS/HPS Meter --
-- Author: Jor @ Deepstrike                --
-- Updated: Wildtide @ Icewatch, 2013      --
---------------------------------------------

--[[

Licensed to the Apache Software Foundation (ASF) under one or more
contributor license agreements.  See the NOTICE file distributed with
this work for additional information regarding copyright ownership.
The ASF licenses this file to You under the Apache License, Version 2.0
(the "License"); you may not use this file except in compliance with
the License.  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

--]]

local toc, data = ...
local AddonId = toc.identifier

---------------
-- Variables --
---------------

local SimpleMeter = {
	version = toc.toc.Version,
	Util = {},
	UI = {
		settings = {
			width = 250,
			name = "Simple Meter",
			alpha = 0.70,
			offset = 4,
			statWidth = 50,
			nameWidth = 250 - 50 * 2 - 4 * 2, -- Width - StatWidth * 2 - Offset * 2
			lineHeight = 10,
		},
		context = nil,
		interface = nil,
		tooltip = nil,
	},
	Event = {},
	Modes = {
		dps = {label = "DPS", desc = "DPS", next = "hps", prev = "htk"},
		hps = {label = "HPS", desc = "HPS", next = "dmg", prev = "dps"},
		dmg = {label = "Dmg", desc = "Damage", next = "heal", prev = "hps"},
		heal = {label = "Heal", desc = "Heals", next = "dtk", prev = "dmg"},
		dtk = {label = "DT", desc = "Damage Taken", next = "htk", prev = "heal"},
		htk = {label = "HR", desc = "Healing Received", next = "dps", prev = "dtk"},
	},
	state = {
		lastUpdate = -1,
		playerId = "",
		encounterIndex = 0,
		empty = {}, -- This will be created after Encounter:new is declared
		all = {},   -- This will be created after Encounter:new is declared
		encounters = {},
		encounterOffset = 0,
		allyOffset = 0,
		enemyOffset = 0,
		viewAbilities = "",
	},
}
SimpleMeter_enabled = true
SimpleMeter_show = true
SimpleMeter_mode = "single"
SimpleMeter_x = 325
SimpleMeter_y = 200
SimpleMeter_expandAllies = "all"
SimpleMeter_expandEnemies = "all"
SimpleMeter_debug = false
SimpleMeter_columns = {"dps", "hps"}
SimpleMeter_sortColumn = 1
SimpleMeter_maxEncounters = 15
SimpleMeter_ignoreAbilities = {
	["Sourcestone Annihilation"] = true,
	["Shocking Cipher"] = true,
	["Deathly Flames"] = true,
}

SimpleMeter.state.playerId = Inspect.Unit.Lookup("player")

-----------------------
-- Class Definitions --
-----------------------

local Encounter = {
	sortEncounter = {},
	sortMode = "dps",
}

local Unit = {}

function Encounter:new()
	local o = {
		inCombat = false,
		combatTime = 0,
		startCombat= 0,
		units = {},
		allies = {},
		enemies = {},
	}
	setmetatable(o, self)
	self.__index = self
	return o
end

function Encounter:StartCombat(time)
	self.inCombat = true
	self.startCombat = time
end

function Encounter:EndCombat(time)
	if self.inCombat then
		self.inCombat = false
		self.combatTime = time - self.startCombat + self.combatTime
	end
end

function Encounter:GetCombatTime(time)
	if self.inCombat and time then
		return time - self.startCombat + self.combatTime
	elseif not self.inCombat then
		return self.combatTime
	else
		return Inspect.Time.Frame() - self.startCombat + self.combatTime
	end
end

function Encounter:GetUnit(id, returnIfPet)
	if not self.units[id] then
		local detail = Inspect.Unit.Detail(id)
		if not detail then
			return nil
		end

		local unit = Unit:new(detail)
		self.units[id] = unit
		if not detail.player and detail.nameSecondary then
			-- May be a pet
			local ownerId = Inspect.Unit.Lookup(id .. ".owner")
			if ownerId then
				unit.owner = self:GetUnit(ownerId)
			end
		end

		if not unit.owner then
			if detail.relation == "friendly" then
				table.insert(self.allies, id)
			else
				table.insert(self.enemies, id)
			end
		end
	end

	local unit = self.units[id]
	if unit.owner then
		if returnIfPet then
			return unit.owner
		else
			return nil
		end
	else
		return unit
	end
end

function Encounter:IsPet(id)
	return self.units[id] and self.units[id].owner
end

function Encounter:Sort(list, mode)
	Encounter.sortEncounter = self
	Encounter.sortMode = mode
	table.sort(list, Encounter.InternalSortFunction)
end

function Encounter.InternalSortFunction(a, b)
	-- Since LUA's sort is unstable, we have to force an order always
	local av = 0
	local bv = 0
	local mode = Encounter.sortMode
	local units = Encounter.sortEncounter.units
	if mode == "dps" or mode == "dmg" then
		av = units[a].damage
		bv = units[b].damage
	elseif mode == "hps" or mode == "heal" then
		av = units[a].heal
		bv = units[b].heal
	elseif mode == "dtk" then
		av = units[a].damageTaken
		bv = units[b].damageTaken
	elseif mode == "htk" then
		av = units[a].healTaken
		bv = units[b].healTaken
	else
		av = units[a].damage
		bv = units[b].damage
	end
	if av == bv then
		av = units[a].name
		bv = units[b].name
		if av == bv then
			return a < b
		else
			return av < bv
		end
	else
		return av > bv
	end
end

function Encounter:BuildCopyText(side, mode, expand)
	local list, copyFrom = {}, {}

	if side == "ally" then
		copyFrom = self.allies
	else
		copyFrom = self.enemies
	end
	if #copyFrom == 0 then
		return ""
	end
	for _,v in pairs(copyFrom) do
		table.insert(list, v)
	end

	self:Sort(list, mode)

	local time = self:GetCombatTime()
	local timeText, totalText, unitText = "", "", ""
	timeText = "Time: " .. SimpleMeter.Util.FormatTime(time)
	if side == "ally" then
		totalText = totalText .. " Ally"
	else
		totalText = totalText .. " Enemy"
	end
	totalText = totalText .. " " .. SimpleMeter.Modes[mode].desc .. ": "

	local total, count = 0, 0
	for _, id in pairs(list) do
		local unit = self.units[id]
		local v = 0
		if mode == "dps" then
			v = unit.damage / time
		elseif mode == "dmg" then
			v = unit.damage
		elseif mode == "hps" then
			v = unit.heal / time
		elseif mode == "heal" then
			v = unit.heal
		elseif mode == "dtk" then
			v = unit.damageTaken
		elseif mode == "htk" then
			v = unit.healTaken
		else
			v = unit.damage / time
		end
		if (expand == "all" and v > 0)
		   or (expand == "top5" and count < 5)
		   or (expand == "self" and id == SimpleMeter.state.playerId) then
			unitText = unitText .. "  " .. unit.name .. " :" .. SimpleMeter.Util.FormatNumber(v)
			count = count + 1
		end
		total = total + v
	end

	local text = timeText .. totalText .. SimpleMeter.Util.FormatNumber(total) .. unitText
	return text
end

function Unit:new(detail)
	local o = {
		name = detail.name,
		calling = detail.calling,
		player = detail.player,
		damage = 0,
		damageTaken = 0,
		heal = 0,
		healTaken = 0,
		owner = false,
		abilities = {},
	}
	setmetatable(o, self)
	self.__index = self
	return o
end

function Unit:AddDamage(ability, damage)
	if SimpleMeter_ignoreAbilities[ability] then return end
	self.damage = self.damage + damage
	local ability = self:GetAbility(ability)
	ability.damage = ability.damage + damage
end

function Unit:AddDamageTaken(ability, damage)
	if SimpleMeter_ignoreAbilities[ability] then return end
	self.damageTaken = self.damageTaken + damage
	local ability = self:GetAbility(ability)
	ability.damageTaken = ability.damageTaken + damage
end

function Unit:AddHeal(ability, heal)
	if SimpleMeter_ignoreAbilities[ability] then return end
	self.heal = self.heal + heal
	local ability = self:GetAbility(ability)
	ability.heal = ability.heal + heal
end

function Unit:AddHealTaken(ability, heal)
	if SimpleMeter_ignoreAbilities[ability] then return end
	self.healTaken = self.healTaken + heal
	local ability = self:GetAbility(ability)
	ability.healTaken = ability.healTaken + heal
end

function Unit:GetAbility(ability)
	if not self.abilities[ability] then
		self.abilities[ability] = {
			damage = 0,
			damageTaken = 0,
			heal = 0,
			healTaken = 0,
		}
	end
	return self.abilities[ability]
end

-----------------
-- GUI Classes --
-----------------

local Interface = {}

local Stats = {
	Line = {
		Name = {},
		Stat = {},
	}
}

function Interface:new()
	local o = {
		state = {},
		headings = {},
		allyStats = {},
		enemyStats = {},
	}
	setmetatable(o, self)
	self.__index = self

	local params = SimpleMeter.UI.settings

	o.frame = UI.CreateFrame("Frame", "SimpleMeter Interface", SimpleMeter.UI.context)
	o.frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", SimpleMeter_x, SimpleMeter_y)
	o.frame:SetLayer(0)
	o.frame:SetBackgroundColor(0, 0, 0, params.alpha)
	o.frame:SetWidth(params.width)
	o.frame:SetVisible(SimpleMeter_show)

	o.title = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Title", parent = o.frame,
		text = params.name, fontSize = 16,
		fontR = 0.7, fontG = 0.7, fontB = 0.7,
		alignTarget = o.frame, alignTo = "TOPCENTER", align = "TOPCENTER", x = 0, y = 0,
	})

	o.prev = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Previous Encounter", parent = o.frame,
		text = "<",
		fontR = 0.7, fontG = 0.7, fontB = 0.7,
		alignTarget = o.frame, alignTo = "TOPLEFT", align = "TOPLEFT", x = params.offset, y = params.offset,
	})

	o.currentBG = SimpleMeter.UI.CreateFrame({
		type = "Frame", name = "Current Encounter Background", parent = o.frame,
		width = 30, hide = true,
		alignTarget = o.prev, alignTo = "TOPRIGHT", align = "TOPLEFT", x = 0, y = 0,
	})

	o.current = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Current Encounter", parent = o.frame,
		text = "-",
		fontR = 0.7, fontG = 0.7, fontB = 0.7,
		alignTarget = o.currentBG, alignTo = "TOPCENTER", align = "TOPCENTER", x = 0, y = 0,
	})

	o.next = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Next Encounter", parent = o.frame,
		text = ">",
		fontR = 0.7, fontG = 0.7, fontB = 0.7,
		alignTarget = o.currentBG, alignTo = "TOPRIGHT", align = "TOPLEFT", x = 0, y = 0,
	})

	o.help = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Help", parent = o.frame,
		text = "?", tooltip = SimpleMeter.GetHelpText(),
		fontR = 0.7, fontG = 0.7, fontB = 0.7,
		alignTarget = o.frame, alignTo = "TOPRIGHT", align = "TOPRIGHT", x = params.offset * -1, y = params.offset,
	})

	o.reset = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "New Session", parent = o.frame,
		text = "N", tooltip = "New Session",
		fontR = 0.7, fontG = 0.7, fontB = 0.7,
		alignTarget = o.help, alignTo = "TOPLEFT", align = "TOPRIGHT", x = params.offset * -1, y = 0,
	})

	o.split = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Split Encounter", parent = o.frame,
		text = "S", tooltip = "Split Encounter",
		fontR = 0.7, fontG = 0.7, fontB = 0.7,
		alignTarget = o.reset, alignTo = "TOPLEFT", align = "TOPRIGHT", x = params.offset * -1, y = 0,
	})

	o.time = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Time", parent = o.frame,
		text = "-",
		width = params.nameWidth,
		alignTarget = o.frame, alignTo = "TOPLEFT", align = "TOPLEFT", x = params.offset, y = o.title:GetHeight() + params.offset,
	})

	o.modeSelect = SimpleMeter.UI.CreateFrame({
		type = "Frame", name = "Mode Select", parent = o.frame,
		layer = 5,
		hide = true,
		r = 0.0, g = 0.0, b = 0.0, a = params.alpha,
	})
	local height = 0
	local width = 0
	for _,v in pairs({"dps", "hps", "dmg", "heal", "dtk", "htk"}) do
		local mode = SimpleMeter.UI.CreateFrame({
			type = "Text", name = "Mode", parent = o.modeSelect,
			text = SimpleMeter.Modes[v].desc,
		})
		mode:SetPoint("TOPLEFT", o.modeSelect, "TOPLEFT", 0, height)
		height = height + mode:GetHeight()
		if mode:GetWidth() > width then
			width = mode:GetWidth()
		end

		mode:EventAttach(Event.UI.Input.Mouse.Left.Down,
			function ()
				o.modeSelect:SetVisible(false)
				o:ChangeHeading(o.modeSelect.heading.index, v)
			end, 
			"MouseLeftDown")
	end
	o.modeSelect:SetHeight(height)
	o.modeSelect:SetWidth(width)

	for i = 1,#SimpleMeter_columns do
		local heading = {
			index = i,
		}
		heading.text = SimpleMeter.UI.CreateFrame({
			type = "Text", name = "Column Label", parent = o.frame,
			text = "-",
			width = params.statWidth - 20,
			alignTarget = o.time, alignTo = "TOPRIGHT", align = "TOPLEFT", x = params.statWidth * (i - 1), y = 0,
		})
		heading.text.parent = heading

		heading.select = SimpleMeter.UI.CreateFrame({
			type = "Texture", name = "Column Select", parent = o.frame,
			source = "SimpleMeter", texture = "down.png",
			alignTarget = heading.text, alignTo = "CENTERRIGHT", align = "CENTERLEFT", x = 0, y = 0,
		})
		heading.select.parent = heading

		heading.text:EventAttach(Event.UI.Input.Mouse.Left.Click,
			function(self, hEvent)
				if SimpleMeter_sortColumn ~= self.parent.index then
					SimpleMeter_sortColumn = self.parent.index
					o:UpdateSort()
					o:Redraw()
				end
			end, "LeftClick")

		heading.text:EventAttach(Event.UI.Input.Mouse.Right.Click,
			function(self, hEvent)
				local mode = SimpleMeter_columns[self.parent.index]
				o:ChangeHeading(self.parent.index, SimpleMeter.Modes[mode].next)
			end, "RightClick")

		heading.select:EventAttach(Event.UI.Input.Mouse.Left.Click,
			function (self, hEvent)
				local heading = self.parent
				if o.modeSelect:GetVisible() and o.modeSelect.heading == heading then
					o.modeSelect:SetVisible(false)
				else
					o.modeSelect.heading = heading
					o.modeSelect:SetPoint("TOPLEFT", heading.text, "BOTTOMLEFT")
					o.modeSelect:SetVisible(true)
				end
			end, "LeftClick")

		table.insert(o.headings, heading)
	end

	o.allyStats = Stats:new("ally", o.frame, o.time, 0, 0.30, 0, params.alpha)
	o.enemyStats = Stats:new("enemy", o.frame, o.allyStats:GetFrame(), 0.30, 0, 0, params.alpha)

	o.copyBoxLabel = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Copybox Label", parent = o.frame,
		text = "Press Ctrl+C to copy to clipboard",
		width = params.width - params.offset * 2,
		hide = true,
		alignTarget = o.enemyStats:GetFrame(), alignTo = "BOTTOMLEFT", align = "TOPLEFT", x = 0, y = 8
	})

	o.copyBox = SimpleMeter.UI.CreateFrame({
		type = "RiftTextfield", name = "Copybox", parent = o.frame,
		text = "-",
		r = 0.3, g = 0.3, b = 0.3, a = params.alpha,
		width = params.width - params.offset * 2,
		hide = true,
		alignTarget = o.copyBoxLabel, alignTo = "BOTTOMLEFT", align = "TOPLEFT", x = 0, y = 0
	})

	o.title:EventAttach(Event.UI.Input.Mouse.Cursor.In, function (self) self:SetFontColor(1, 1, 1) end, "MouseIn")
	o.title:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function (self) self:SetFontColor(0.8, 0.8, 0.8) end, "MouseOut")

	o.title:EventAttach(Event.UI.Input.Mouse.Left.Down, 
		function ()
			o.state.mouseDown = true
			local mouse = Inspect.Mouse()
			o.state.startX = o.frame:GetLeft()
			o.state.startY = o.frame:GetTop()
			o.state.mouseStartX = mouse.x
			o.state.mouseStartY = mouse.y
			o.frame:SetBackgroundColor(0.2, 0.2, 0.2, params.alpha)
		end, "LeftDown")

	o.title:EventAttach(Event.UI.Input.Mouse.Cursor.Move, 
		function ()
			if o.state.mouseDown then
				local mouse = Inspect.Mouse()
				SimpleMeter_x = mouse.x - o.state.mouseStartX + o.state.startX
				SimpleMeter_y = mouse.y - o.state.mouseStartY + o.state.startY
				o.frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", SimpleMeter_x, SimpleMeter_y)
			end
		end, "MouseMove")

	o.title:EventAttach(Event.UI.Input.Mouse.Left.Up, 
		function ()
			if o.state.mouseDown then
				o.state.mouseDown = false
				o.frame:SetBackgroundColor(0, 0, 0, params.alpha)
			end
		end, "LeftUp")

	o.current:EventAttach(Event.UI.Input.Mouse.Left.Click, 
		function ()
			if SimpleMeter_mode == "all" then
				SimpleMeter_mode = "single"
			else
				SimpleMeter_mode = "all"
				SimpleMeter.state.encounterOffset = 0
			end
			SimpleMeter.UI.interface:ChangeEncounter()
		end, "LeftClick")

	o.reset:EventAttach(Event.UI.Input.Mouse.Left.Click, 
		function ()
			SimpleMeter.ResetState()
			SimpleMeter.UI.interface:ChangeEncounter()
		end, "LeftClick")

	o.split:EventAttach(Event.UI.Input.Mouse.Left.Click, 
		function ()
			if SimpleMeter.state.all.inCombat then
				SimpleMeter.Event.EndCombat()
				SimpleMeter.Event.StartCombat()
			end
		end, "LeftClick")

	o.prev.OnLeftClick = 
		function()
			if SimpleMeter_mode == "single" then
				if SimpleMeter.state.encounterOffset < #SimpleMeter.state.encounters - 1 then
					SimpleMeter.state.encounterOffset = SimpleMeter.state.encounterOffset + 1
					SimpleMeter.UI.interface:ChangeEncounter()
				end
			end
		end

	o.next.OnLeftClick = 
		function()
			if SimpleMeter_mode == "single" then
				if SimpleMeter.state.encounterOffset > 0 then
					SimpleMeter.state.encounterOffset = SimpleMeter.state.encounterOffset - 1
					SimpleMeter.UI.interface:ChangeEncounter()
				end
			end
		end
			
	o.prev:EventAttach(Event.UI.Input.Mouse.Left.Click, o.prev.OnLeftClick, "LeftClick") 
	o.next:EventAttach(Event.UI.Input.Mouse.Left.Click, o.next.OnLeftClick, "LeftClick")		
	o.current:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function () o.prev.Event.LeftClick() end, "WheelBack")
	o.current:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function () o.next.Event.LeftClick() end, "WheelBack")

	o.copyBox:EventAttach(Event.UI.Input.Key.Focus.Loss, 
		function ()
			o.copyBox:SetVisible(false)
			o.copyBoxLabel:SetVisible(false)
			o:Resize()
		end, "FocusLoss")
	
	return o
end

function Interface:Show()
	self.frame:SetVisible(true)
end

function Interface:Hide()
	self.frame:SetVisible(false)
end

function Interface:Toggle()
	self.frame:SetVisible(not self.frame:GetVisible())
end

function Interface:ChangeHeading(index, mode)
	SimpleMeter_columns[index] = mode
	self:UpdateSort()
	self:Redraw()
end

function Interface:UpdateSort()
	for i = 1, #SimpleMeter_columns do
		local mode = SimpleMeter_columns[i]
		self.headings[i].text:SetText(SimpleMeter.Modes[mode].label)
		if SimpleMeter_sortColumn == i then
			self.headings[i].text:SetFontColor(1.0, 1.0, 1.0)
			self.headings[i].select:SetAlpha(1.0)
		else
			self.headings[i].text:SetFontColor(0.7, 0.7, 0.7)
			self.headings[i].select:SetAlpha(0.7)
		end
	end
end

function Interface:SetCopyText(text)
	if text and text:len() > 0 then
		self.copyBox:SetText(text)
		self.copyBoxLabel:SetVisible(true)
		self.copyBox:SetVisible(true)
		self.copyBox:SetKeyFocus(true)
		self.copyBox:SetSelection(0, text:len())
	else
		self.copyBoxLabel:SetVisible(false)
		self.copyBox:SetVisible(false)
	end
	self:Resize()
end

function Interface:Resize()
	local height = 0
	local h = 0

	local params = SimpleMeter.UI.settings
	height = height + self.title:GetHeight() + params.offset + self.time:GetHeight() + params.offset

	self.allyStats:Resize()
	height = height + self.allyStats.height

	self.enemyStats:Resize()
	height = height + self.enemyStats.height

	if self.copyBox:GetVisible() then
		height = height + self.copyBoxLabel:GetHeight() + 8
		height = height + self.copyBox:GetHeight()
	end

	self.frame:SetHeight(height)
end

function Interface:Init()
	self.frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", SimpleMeter_x, SimpleMeter_y)
	if SimpleMeter_enabled then
		self.frame:SetVisible(SimpleMeter_show)
	else
		self.frame:SetVisible(false)
	end
	self:UpdateSort()
	self:ChangeEncounter()
end

function Interface:Redraw()
	local encounter = encounter()
	local time = encounter:GetCombatTime()

	self.time:SetText("Time: " .. SimpleMeter.Util.FormatTime(time));

	local ally = SimpleMeter.UI.GetStats("ally", time, SimpleMeter_expandAllies, SimpleMeter.state.allyOffset, SimpleMeter_columns)
	self.allyStats:Set(ally)

	local enemy = SimpleMeter.UI.GetStats("enemy", time, SimpleMeter_expandEnemies, SimpleMeter.state.enemyOffset, SimpleMeter_columns)
	self.enemyStats:Set(enemy)

	self:Resize()
end

function Interface:ShowAbilities(side, mode, expand, index)
	if SimpleMeter_debug then print("Show abilities") end

	local encounter = encounter()

	if side == "enemy" then
		list = encounter.enemies
		if expand == "top5"  then
			index = index + SimpleMeter.state.enemyOffset
		end
		id = "e"
	else
		list = encounter.allies
		if expand == "top5"  then
			index = index + SimpleMeter.state.allyOffset
		end
		id = "a"
	end

	id = id .. index
	if SimpleMeter.state.viewAbilities == id then
		-- Abilities are already being shown
		return
	else
		SimpleMeter.state.viewAbilities = id
	end

	if index <= 0 or index > #list then
		SimpleMeter.UI.tooltip:Hide()
		return
	end

	local units = encounter.units
	local unit = false
	if expand == "all" or expand == "top5" then
		unit = units[list[index]]
	elseif expand == "self" then
		unit = units[SimpleMeter.state.playerId]
	else
		unit = false
	end

	if not unit or (mode == "dps" and unit.damage == 0) or (mode == "hps" and unit.heal == 0) then
		SimpleMeter.UI.tooltip:Hide()
		return
	end

	local names = {}
	local value
	local total

	for k,_ in pairs(unit.abilities) do
		table.insert(names, k)
	end
	if mode == "hps" or mode == "heal" then
		table.sort(names, function(a, b) return unit.abilities[a].heal > unit.abilities[b].heal end)
	elseif mode == "dtk" then
		table.sort(names, function(a, b) return unit.abilities[a].damageTaken > unit.abilities[b].damageTaken end)
	elseif mode == "htk" then
		table.sort(names, function(a, b) return unit.abilities[a].healTaken > unit.abilities[b].healTaken end)
	else
		table.sort(names, function(a, b) return unit.abilities[a].damage > unit.abilities[b].damage end)
	end

	local text = unit.name .. "'s " .. SimpleMeter.Modes[mode].desc .. ":\n\n"
	for _,name in pairs(names) do
		if mode == "hps" or mode == "heal" then
			value = unit.abilities[name].heal
			total = unit.heal
		elseif mode == "dtk" then
			value = unit.abilities[name].damageTaken
			total = unit.damageTaken
		elseif mode == "htk" then
			value = unit.abilities[name].healTaken
			total = unit.healTaken
		else
			value = unit.abilities[name].damage
			total = unit.damage
		end
		if value > 0 then
			text = text .. string.format("%.0f%% %s\n", value / total * 100, name)
		end
	end
	SimpleMeter.UI.tooltip:Show(text)
end

function Interface:HideAbilities()
	SimpleMeter.state.viewAbilities = ""
	SimpleMeter.UI.tooltip:Hide()
end

function Interface:ChangeEncounter()
	if SimpleMeter_mode == "all" then
		self.current:SetText("All")
	elseif SimpleMeter.state.encounterIndex == 0 then
		self.current:SetText("-/-")
	else
		self.current:SetText(string.format("%2d/%d", SimpleMeter.state.encounterIndex - SimpleMeter.state.encounterOffset, #SimpleMeter.state.encounters ))
	end
	self:Redraw()
end

function Stats:new(side, parent, alignTarget, r, g, b, a)
	local o = {
		side = side,
		lines = {},
		height = 0,
	}
	setmetatable(o, self)
	self.__index = self

	local params = SimpleMeter.UI.settings

	o.frame = SimpleMeter.UI.CreateFrame({
		type = "Frame", name = "Stats Panel", parent = parent,
		width = params.nameWidth + params.statWidth * 2,
		r = r, g = g, b = b, a = a,
		alignTarget = alignTarget, alignTo = "BOTTOMLEFT", align = "TOPLEFT", x = 0, y = 0,
	})

	return o
end

function Stats:GetExpand()
	if self.side == "ally" then
		return SimpleMeter_expandAllies
	else
		return SimpleMeter_expandEnemies
	end
end

function Stats:SetExpand(expand)
	if self.side == "ally" then
		SimpleMeter_expandAllies = expand
		if SimpleMeter_debug then print("Expand allies set to " .. expand) end
	else
		SimpleMeter_expandEnemies = expand
		if SimpleMeter_debug then print("Expand enemies set to " .. expand) end
	end
end

function Stats:GetFrame()
	return self.frame
end

function Stats:Set(stats)
	for i = 1,#stats,1 do
		if not self.lines[i] then
			if i == 1 then
				self.lines[i] = Stats.Line:new("ally", i, self, self.frame, "TOPLEFT", 0)
			else
				self.lines[i] = Stats.Line:new("enemy", i, self, self.lines[i - 1].name, "BOTTOMLEFT", -4)
			end
		end
		self.lines[i]:Set(stats[i])
		self.lines[i]:Show()
	end
	for i = #stats + 1, #self.lines, 1 do
		self.lines[i]:Hide()
	end
end

function Stats:Resize()
	local height = 4
	for i = 1, #self.lines, 1 do
		if self.lines[i].visible then
			height = height + self.lines[i].height - 4
		end
	end
	self.frame:SetHeight(height)
	self.height = height
end

function Stats.Line:new(side, index, parent, alignTarget, alignTo, y)
	local o = {
		visible = false,
		height = 0,
		side = side,
		index = index,
		columns = {},
	}
	setmetatable(o, self)
	self.__index = self

	local params = SimpleMeter.UI.settings

	o.name = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Name", parent = parent.frame,
		text = "-",
		width = params.nameWidth,
		hide = true,
		alignTarget = alignTarget, alignTo = alignTo, align = "TOPLEFT", x = 0, y = y,
	})
	o.name.parent = parent
	o.name.line = o
	
	o.name:EventAttach(Event.UI.Input.Mouse.Left.Click, Stats.Line.Name.LeftClick, "LeftClick") 
	o.name:EventAttach(Event.UI.Input.Mouse.Wheel.Back, Stats.Line.Name.WheelBack, "WheelBack") 
	o.name:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, Stats.Line.Name.WheelForward, "WheelForward") 

	for i = 1,#SimpleMeter_columns do
		o.columns[i] = SimpleMeter.UI.CreateFrame({
			type = "Text", name = "DPS", parent = parent.frame,
			text = "-",
			width = params.statWidth,
			alignTarget = o.name, alignTo = "TOPRIGHT", align = "TOPLEFT", x = params.statWidth * (i - 1), y = 0,
		})
		o.columns[i].parent = parent
		o.columns[i].line = o
		o.columns[i].index = i

		o.columns[i]:EventAttach(Event.UI.Input.Mouse.Left.Click, Stats.Line.Stat.LeftClick, "LeftClick") 
		o.columns[i]:EventAttach(Event.UI.Input.Mouse.Middle.Click, Stats.Line.Stat.MiddleClick, "MiddleClick") 
		o.columns[i]:EventAttach(Event.UI.Input.Mouse.Cursor.Move, Stats.Line.Stat.MouseMove, "MouseMove") 
		o.columns[i]:EventAttach(Event.UI.Input.Mouse.Cursor.Out, Stats.Line.Stat.MouseOut, "MouseOut") 
		o.columns[i]:EventAttach(Event.UI.Input.Mouse.Wheel.Back, Stats.Line.Stat.WheelBack, "WheelBack") 
		o.columns[i]:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, Stats.Line.Stat.WheelForward, "WheelForward") 

	end

	o.height = o.name:GetHeight()
	
	return o
end

function Stats.Line:Show()
	self.name:SetVisible(true)
	for i = 1,#self.columns do
		self.columns[i]:SetVisible(true)
	end
	self.visible = true
end

function Stats.Line:Hide()
	self.name:SetVisible(false)
	for i = 1,#self.columns do
		self.columns[i]:SetVisible(false)
	end
	self.visible = false
end

function Stats.Line:Set(stat)
	if stat.name == nil then
		self.name:SetText("-")
	else
		self.name:SetText(stat.name)
	end

	for i = 1,#self.columns do
		local mode = SimpleMeter_columns[i]
		if stat[mode] == nil then
			self.columns[i]:SetText("-")
		else
			self.columns[i]:SetText(SimpleMeter.Util.FormatNumber(stat[mode]))
		end
	end

	if stat.r then
		self.name:SetFontColor(stat.r, stat.g, stat.b)
	else
		self.name:SetFontColor(1.0, 1.0, 1.0)
	end
end

function Stats.Line.Name:LeftClick()
	local expand = self.parent:GetExpand()
	local side = self.parent.side
	if expand == "all" then
		expand = "top5"
	elseif expand == "top5" and side == "ally" then
		expand = "self"
	elseif expand == "top5" or expand == "self" then
		expand = "total"
	else
		expand = "all"
	end
	self.parent:SetExpand(expand)

	SimpleMeter.UI.interface:Redraw()
end
function Stats.Line.Name:WheelBack()
	SimpleMeter.UI.Scroll(self.parent.side, 1)
end
function Stats.Line.Name:WheelForward()
	SimpleMeter.UI.Scroll(self.parent.side, -1)
end

function Stats.Line.Stat:LeftClick()
	SimpleMeter.UI.interface:SetCopyText(encounter():BuildCopyText(self.parent.side, SimpleMeter_columns[self.index], "all"))
end
function Stats.Line.Stat:MiddleClick()
	SimpleMeter.UI.interface:SetCopyText(encounter():BuildCopyText(self.parent.side, SimpleMeter_columns[self.index], self.parent:GetExpand()))
end
function Stats.Line.Stat:MouseMove()
	SimpleMeter.UI.interface:ShowAbilities(self.parent.side, SimpleMeter_columns[self.index], self.parent:GetExpand(), self.line.index)
end
function Stats.Line.Stat:MouseOut()
	SimpleMeter.UI.interface:HideAbilities()
end
function Stats.Line.Stat:WheelBack()
	SimpleMeter.UI.Scroll(self.parent.side, 1)
end
function Stats.Line.Stat:WheelForward()
	SimpleMeter.UI.Scroll(self.parent.side, -1)
end

local Tooltip = {}
function Tooltip:new()
	local o = {}
	setmetatable(o, self)
	self.__index = self

	local params = SimpleMeter.UI.settings

	o.frame = SimpleMeter.UI.CreateFrame({
		type = "Text", name = "Tooltip", parent = SimpleMeter.UI.context,
		layer = 5,
		text = "-", fontSize = 13, wordwrap = true,
		hide = true,
	})

	o.background = SimpleMeter.UI.CreateFrame({
		type = "Frame", name = "Tooltip Background", parent = SimpleMeter.UI.context,
		layer = 4,
		r = 0, g = 0, b = 0, a = params.alpha,
		border = 2, borderR = 0.7, borderG = 0.7, borderB = 0.4, borderA = 0.9,
		hide = true,
		padTarget = o.frame, pad = 14 
	})

	return o
end

function Tooltip:Show(text, anchor)
	self.frame:ClearAll()
	self.frame:SetText(text)
	self.frame:SetWidth(210)

	if anchor then
		self.frame:SetPoint("TOPLEFT", anchorTo, "BOTTOMRIGHT", 40, 40)
	else
		self.frame:SetPoint("BOTTOMRIGHT", UIParent, "BOTTOMRIGHT", -40, -120)
	end

	self.background:SetVisible(true)
	self.frame:SetVisible(true)
end

function Tooltip:Hide()
	self.frame:SetVisible(false)
	self.background:SetVisible(false)
end

------------------------
-- State Lookup Method --
-------------------------

function encounter()
	if SimpleMeter_mode == "all" then
		return SimpleMeter.state.all
	elseif SimpleMeter.state.encounterIndex == 0 then
		return SimpleMeter.state.empty
	else
		return SimpleMeter.state.encounters[SimpleMeter.state.encounterIndex - SimpleMeter.state.encounterOffset]
	end
end

--------------------
-- Initialization --
--------------------

SimpleMeter.state.all = Encounter:new()
SimpleMeter.state.empty = Encounter:new()

function SimpleMeter.Init(hEvent, addon)
	if addon == "SimpleMeter" then
		-- Initialize the UI
		SimpleMeter.UI.context = UI.CreateContext("SimpleMeter")

		SimpleMeter.UI.interface = Interface:new()
		SimpleMeter.UI.tooltip = Tooltip:new()

		SimpleMeter.UI.interface:Init()

		print(SimpleMeter.version .. " Loaded.  Type /sm to see a full list of commands.")
	end
end

function SimpleMeter.ResetState()
	local inCombat = SimpleMeter.state.all.inCombat

	SimpleMeter.state.encounterIndex = 0 
	SimpleMeter.state.all = Encounter:new()
	SimpleMeter.state.encounters = {}

	SimpleMeter.state.encounterOffset = 0
	SimpleMeter.state.allyOffset = 0
	SimpleMeter.state.enemyOffset = 0
	SimpleMeter.state.viewAbilities = ""

	if inCombat then
		SimpleMeter.Event.StartCombat()
	end
end

----------
-- Util --
----------

function SimpleMeter.Util.FormatTime(time)
	local hour = math.floor(time / 60 / 60)
	local min = math.floor(time / 60) % 60
	local sec = time % 60

	if hour > 0 then
		return string.format("%02d:%02d:%02d", hour, min, sec)
	else
		return string.format("%02d:%02d", min, sec)
	end
end

function SimpleMeter.Util.FormatNumber(number)
	if number < 10000 then
		return string.format("%d", number)
	elseif number < 100000 then
		return string.format("%4.1fk", number / 1000)
	elseif number < 1000000 then
		return string.format("%dk", number / 1000)
	elseif number < 10000000 then
		return string.format("%4.2fm", number / 1000000)
	elseif number < 100000000 then
		return string.format("%4.1fm", number / 1000000)
	else
		return string.format("%dm", number / 1000000)
	end
end

--------------------
--------------------
-- Slash Commands --
--------------------

function SimpleMeter.SlashHandler(hEvent, args)
	local cmd = args
	if cmd == "version" then
		print("SimpleMeter version "..SimpleMeter.version)
		return
	elseif cmd == "debug" then
		SimpleMeter_debug = not SimpleMeter_debug
		if SimpleMeter_debug then
			print("Debugging enabled for SimpleMeter")
		else
			print("Debugging disabled for SimpleMeter")
		end
		return
	elseif cmd == "show" then
		SimpleMeter.UI.interface:Show()
		return
	elseif cmd == "hide" then
		SimpleMeter.UI.interface:Hide()
		return
	elseif cmd == "toggle" then
		SimpleMeter.UI.interface:Toggle()
		return
	elseif cmd == "enable" then
		SimpleMeter_enabled = true
		print("enabled")
		if SimpleMeter_show then
			SimpleMeter.UI.interface:Show()
		end
		return
	elseif cmd == "disable" then
		SimpleMeter_enabled = false
		print("disabled")
		SimpleMeter.UI.interface:Hide()
		return
	elseif cmd == "reset" then
		SimpleMeter_enabled = true
		SimpleMeter_show = true
		SimpleMeter_x = 325
		SimpleMeter_y = 200
		SimpleMeter_debug = false
		SimpleMeter.UI.interface:Init()
		return
	end
	print("SimpleMeter is an easy-to-use DPS/HPS meter that shows damage and healing for both allies and enemies.")
	print("/sm version - print the version")
	print("/sm toggle - show/hide Simple Meter")
	print("/sm show/hide - show/hide Simple Meter")
	print("/sm enable/disable - enable/disable Simple Meter")
	print("/sm reset - Reset Simple Meter to default position")
end

function SimpleMeter.GetHelpText()
	text = ""
	text = text .. "Click and drag the title bar to move the window.\n\n"
	text = text .. "Click the DPS or HPS labels to change sorting.  Change what stat you want to show by selecting it from the drop-down\n\n"
	text = text .. "Hover over DPS or HPS numbers to see a breakdown of the damage or healing by ability\n\n"
	text = text .. "By clicking the green or red sections you can switch between showing full stats, top 5, only yourself, or only the totals.\n\n"
	text = text .. "To export ally/enemy DPS/HPS, click the DPS/HPS numbers in either the ally or enemy sections.  This will bring up a textbox that you can copy the export from.\n\n"
	text = text .. "To switch between encounter and session modes, click on the encounter indicator in the upper left of the window.  Use the 'N' to start a new session, and the 'S' to split an encounter\n\n"
	text = text .. "Make a macro with /sm toggle to show/hide the window.  Type /sm to see a full list of commands.\n"
	return text
end

------------------
-- GUI Commands --
------------------

function SimpleMeter.UI.CreateFrame(p)
	local frame = UI.CreateFrame(p.type, p.name, p.parent)
	if p.hide then
		frame:SetVisible(false)
	end
	if p.layer then
		frame:SetLayer(p.layer)
	end

	if p.type == "Text" then
		if p.fontSize then
			frame:SetFontSize(p.fontSize)
		else
			frame:SetFontSize(12)
		end
	elseif p.type == "Texture" then
		frame:SetTexture(p.source, p.texture)
	end

	if p.fontR then
		frame:SetFontColor(p.fontR, p.fontG, p.fontB)
	end
	if p.a then
		frame:SetBackgroundColor(p.r, p.g, p.b, p.a)
	end

	if p.text then
		frame:SetText(p.text)
	end
	if p.wordwrap then
		frame:SetWordwrap(p.wordwrap)
	end

	if p.width then
		frame:SetWidth(p.width)
	end

	if p.height then
		frame:SetHeight(p.height)
	end

	if p.align then
		frame:SetPoint(p.align, p.alignTarget, p.alignTo, p.x, p.y)
	end

	if p.padTarget then
		frame:SetPoint("TOPLEFT", p.padTarget, "TOPLEFT", p.pad * -1, p.pad * -1)
		frame:SetPoint("BOTTOMRIGHT", p.padTarget, "BOTTOMRIGHT", p.pad, p.pad)
	end

	if p.border then
		frame.borderTop = UI.CreateFrame("Frame", p.name .. "_btop", frame)
		frame.borderTop:SetLayer(frame:GetLayer() + 1)
		frame.borderTop:SetBackgroundColor(p.borderR, p.borderG, p.borderB, p.borderA)
		frame.borderTop:SetPoint("TOPLEFT", frame, "TOPLEFT")
		frame.borderTop:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 0, p.border)

		frame.borderBot = UI.CreateFrame("Frame", p.name .. "_bbot", frame)
		frame.borderBot:SetLayer(frame:GetLayer() + 1)
		frame.borderBot:SetBackgroundColor(p.borderR, p.borderG, p.borderB, p.borderA)
		frame.borderBot:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", 0, p.border * -1)
		frame.borderBot:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT")

		frame.borderRt = UI.CreateFrame("Frame", p.name .. "_brt", frame)
		frame.borderRt:SetLayer(frame:GetLayer() + 1)
		frame.borderRt:SetBackgroundColor(p.borderR, p.borderG, p.borderB, p.borderA)
		frame.borderRt:SetPoint("TOPLEFT", frame, "TOPLEFT")
		frame.borderRt:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", p.border, 0)

		frame.borderLt = UI.CreateFrame("Frame", p.name .. "_blt", frame)
		frame.borderLt:SetLayer(frame:GetLayer() + 1)
		frame.borderLt:SetBackgroundColor(p.borderR, p.borderG, p.borderB, p.borderA)
		frame.borderLt:SetPoint("TOPLEFT", frame, "TOPRIGHT", p.border * -1, 0)
		frame.borderLt:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT")
	end

	if p.tooltip then
		frame:EventAttach(Event.UI.Input.Mouse.Cursor.In, function () SimpleMeter.UI.tooltip:Show(p.tooltip) end, "MouseIn")
		frame:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function () SimpleMeter.UI.tooltip:Hide() end, "MouseOut")
	end

	return frame;
end

function SimpleMeter.UI.Scroll(side, num)
	if side == "ally" then
		expand = SimpleMeter_expandAllies
	else
		expand = SimpleMeter_expandEnemies
	end
	if expand ~= "top5" then return end

	local encounter = encounter()
	if side == "ally" then
		local offset = SimpleMeter.state.allyOffset + num
		if offset < 0 or #encounter.allies < 5 then
			offset = 0
		elseif offset > #encounter.allies - 5 then
			offset = #encounter.allies - 5
		end
		SimpleMeter.state.allyOffset = offset
	else
		local offset = SimpleMeter.state.enemyOffset + num
		if offset < 0 or #encounter.enemies < 5 then
			offset = 0
		elseif offset > #encounter.enemies - 5 then
			offset = #encounter.enemies - 5
		end
		SimpleMeter.state.enemyOffset = offset
	end

	SimpleMeter.UI.interface:Redraw()
end

function SimpleMeter.UI.Update(hEvent)
	if not SimpleMeter_enabled then return end
	if not encounter().inCombat then return end

	local now = Inspect.Time.Frame()

	-- Update 5 times a second
	if (now - SimpleMeter.state.lastUpdate) > 0.2 then
		SimpleMeter.state.lastUpdate = now
	else
		return
	end

	SimpleMeter.UI.interface:Redraw()
end

function SimpleMeter.UI.GetStats(side, time, expand, offset, columns)

	local encounter = encounter()

	local unitList = {}
	if side == "enemy" then
		unitList = encounter.enemies
	else
		unitList = encounter.allies
	end

	if #unitList > 1 and (expand == "top5" or expand == "all") then
		encounter:Sort(unitList, SimpleMeter_columns[SimpleMeter_sortColumn])
	end

	if expand ~= "top5" or offset < 0 or offset > #unitList - 5 then
		offset = 0
	end

	local stats = {}
	local totalDamage, totalDamageTaken, totalHeal, totalHealTaken = 0, 0, 0, 0
	for _, id in pairs(unitList) do
		local unit = encounter.units[id]
		if expand == "all"
		   or (expand == "top5" and #stats < 5)
		   or (expand == "self" and id == SimpleMeter.state.playerId) then
		   	if offset > 0 then
				offset = offset - 1
			else
				local stat = {
					name = unit.name,
				}
				for i=1,#columns do
					if columns[i] == "dps" then
						stat.dps = unit.damage / time
					elseif columns[i] == "hps" then
						stat.hps = unit.heal / time
					elseif columns[i] == "dmg" then
						stat.dmg = unit.damage
					elseif columns[i] == "heal" then
						stat.heal = unit.heal
					elseif columns[i] == "dtk" then
						stat.dtk = unit.damageTaken
					elseif columns[i] == "htk" then
						stat.htk = unit.healTaken
					end
				end
				if unit.player and id ~= SimpleMeter.state.playerId then
					if unit.calling == "mage" then
						stat.r, stat.g, stat.b = 0.65, 0.50, 1.00 -- 2.15
					elseif unit.calling == "rogue" then
						stat.r, stat.g, stat.b = 1.00, 0.80, 0.25 -- 2.05
					elseif unit.calling == "cleric" then
						stat.r, stat.g, stat.b = 0.60, 1.00, 0.25 -- 1.85
					elseif unit.calling == "warrior" then
						stat.r, stat.g, stat.b = 1.00, 0.35, 0.45 -- 1.80
					end
				end
				table.insert(stats, stat)
			end
		end
		totalDamage = totalDamage + unit.damage
		totalDamageTaken = totalDamageTaken + unit.damageTaken
		totalHeal = totalHeal + unit.heal
		totalHealTaken = totalHealTaken + unit.healTaken
	end

	if expand == "top5" then
		while #stats < 5 do
			local stat = {
				name = nil,
			}
			for i=1,#columns do
				stat[columns[i]] = nil
			end
			table.insert(stats, stat)
		end
	end

	if (expand == "all" and #unitList > 1)
	   or expand == "top5"
	   or (expand == "total" and #unitList > 0) then
	   	local stat = {
			name = "Total",
			r = 0.7,
			g = 0.7,
			b = 0.7,
		}
		for i=1,#columns do
			if #unitList > 0 then
				if columns[i] == "dps" then
					stat.dps = totalDamage / time
				elseif columns[i] == "hps" then
					stat.hps = totalHeal / time
				elseif columns[i] == "dmg" then
					stat.dmg = totalDamage
				elseif columns[i] == "heal" then
					stat.heal = totalHeal
				elseif columns[i] == "dtk" then
					stat.dtk = totalDamageTaken
				elseif columns[i] == "htk" then
					stat.htk = totalHealTaken
				end
			else
				stat[columns[i]] = nil
			end
		end
		table.insert(stats, stat)
	end

	if #stats == 0 then
		local stat = {}
		if expand == "total" then
			stat.name = "Total"
		elseif expand == "self" then
			stat.name = "Self"
		else
			stat.name = nil
		end
		for i=1,#columns do
			stat[columns[i]] = nil
		end
		table.insert(stats, stat)
	end

	return stats
end

--------------------
-- Event Handlers --
--------------------

function SimpleMeter.Event.StartCombat(hEvent)
	if not SimpleMeter_enabled then return end
	if SimpleMeter_debug then print("Starting combat") end
	local now = Inspect.Time.Frame()

	SimpleMeter.state.lastUpdate = -1
	SimpleMeter.state.allyOffset = 0
	SimpleMeter.state.enemyOffset = 0

	SimpleMeter.state.encounterIndex = SimpleMeter.state.encounterIndex + 1
	if SimpleMeter.state.encounterOffset > 0 then
		SimpleMeter.state.encounterOffset = SimpleMeter.state.encounterOffset  + 1
	end
	while SimpleMeter.state.encounterIndex > SimpleMeter_maxEncounters do
		table.remove(SimpleMeter.state.encounters, 1)
		SimpleMeter.state.encounterIndex = SimpleMeter.state.encounterIndex - 1
		if SimpleMeter.state.encounterOffset > 0 then
			SimpleMeter.state.encounterOffset = SimpleMeter.state.encounterOffset  - 1
		end
	end

	local encounter = Encounter:new()
	SimpleMeter.state.encounters[SimpleMeter.state.encounterIndex] = encounter

	encounter:StartCombat(now)
	SimpleMeter.state.all:StartCombat(now)

	SimpleMeter.UI.interface:ChangeEncounter()
end

function SimpleMeter.Event.EndCombat(hEvent)
	if SimpleMeter_debug then print("Ending combat") end
	if not SimpleMeter_enabled and not SimpleMeter.state.all.inCombat then return end
	local now = Inspect.Time.Frame()

	SimpleMeter.state.encounters[SimpleMeter.state.encounterIndex]:EndCombat(now)
	SimpleMeter.state.all:EndCombat(now)

	SimpleMeter.UI.interface:ChangeEncounter()
end

function SimpleMeter.Event.Damage(hEvent, params)
	if not SimpleMeter_enabled then return end
	if SimpleMeter_debug then print(params.damage .. " damage from " .. params.caster) end
	if not SimpleMeter.state.all.inCombat then return end
	if not params.damage then return end

	local caster = params.caster
	local target = params.target
	local currentCaster = SimpleMeter.state.encounters[SimpleMeter.state.encounterIndex]:GetUnit(caster, true)
	local currentTarget = SimpleMeter.state.encounters[SimpleMeter.state.encounterIndex]:GetUnit(target, false)
	local allCaster = SimpleMeter.state.all:GetUnit(caster, true)
	local allTarget = SimpleMeter.state.all:GetUnit(target, false)

	local ability = params.abilityName
	if SimpleMeter.state.all:IsPet(caster) then
		ability = ability .. " (Pet)"
	end

	if currentCaster then
		currentCaster:AddDamage(ability, params.damage)
	end
	if currentTarget then
		currentTarget:AddDamageTaken(ability, params.damage)
	end
	if allCaster then
		allCaster:AddDamage(ability, params.damage)
	end
	if allTarget then
		allTarget:AddDamageTaken(ability, params.damage)
	end
end

function SimpleMeter.Event.Heal(hEvent, params)
	if not SimpleMeter_enabled then return end
	if SimpleMeter_debug then print(params.heal .. " heal from " .. params.caster) end
	if not SimpleMeter.state.all.inCombat then return end
	if not params.heal then return end

	local caster = params.caster
	local target = params.target
	local currentCaster = SimpleMeter.state.encounters[SimpleMeter.state.encounterIndex]:GetUnit(caster, true)
	local currentTarget = SimpleMeter.state.encounters[SimpleMeter.state.encounterIndex]:GetUnit(target, false)
	local allCaster = SimpleMeter.state.all:GetUnit(caster, true)
	local allTarget = SimpleMeter.state.all:GetUnit(target, false)

	local ability = params.abilityName
	if SimpleMeter.state.all:IsPet(caster) then
		ability = ability .. " (Pet)"
	end

	if currentCaster then
		currentCaster:AddHeal(ability, params.heal)
	end
	if currentTarget then
		currentTarget:AddHealTaken(ability, params.heal)
	end
	if allCaster then
		allCaster:AddHeal(ability, params.heal)
	end
	if allTarget then
		allTarget:AddHealTaken(ability, params.heal)
	end
end

---------------------
-- Register Events --
---------------------
Command.Event.Attach(Event.Addon.Load.End, SimpleMeter.Init, "Initital Setup")
Command.Event.Attach(Command.Slash.Register("sm"), SimpleMeter.SlashHandler, "Slash Command")
Command.Event.Attach(Event.System.Update.Begin, SimpleMeter.UI.Update, "Update UI")
Command.Event.Attach(Event.System.Secure.Enter, SimpleMeter.Event.StartCombat, "Start Combat")
Command.Event.Attach(Event.System.Secure.Leave, SimpleMeter.Event.EndCombat, "End Combat")
Command.Event.Attach(Event.Combat.Damage, SimpleMeter.Event.Damage, "Handle Damage")
Command.Event.Attach(Event.Combat.Heal, SimpleMeter.Event.Heal, "Handle Heals")
